/*
 * Copyright (C) 2003-2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/*
 * This file is based in part on mp3tech.c and mp3tech.h, Copyright (C)
 * 2000-2001 Cedric Tefft <cedric@earthling.net>, which in turn is based
 * in part on
 *
 *	* MP3Info 0.5 by Ricardo Cerqueira <rmc@rccn.net>
 *	* MP3Stat 0.9 by Ed Sweetman <safemode@voicenet.com> and 
 *			 Johannes Overmann <overmann@iname.com>
 */

#include "server.h"

/* MIN_CONSEC_GOOD_FRAMES defines how many consecutive valid MP3 frames
   we need to see before we decide we are looking at a real MP3 file */
#define MIN_CONSEC_GOOD_FRAMES 3
#define FRAME_HEADER_SIZE 4
#define MIN_FRAME_SIZE 21

typedef struct {
	unsigned long	sync;
	unsigned int	version;
	unsigned int	layer;
	unsigned int	crc;
	unsigned int	bitrate;
	unsigned int	freq;
	unsigned int	padding;
	unsigned int	extension;
	unsigned int	mode;
	unsigned int	mode_extension;
	unsigned int	copyright;
	unsigned int	original;
	unsigned int	emphasis;
} mp3header;

typedef struct {
	char title[31];
	char artist[31];
	char album[31];
	char year[5];
	char comment[31];
	unsigned char track[1];
	unsigned char genre[1];
} id3tag;

typedef struct {
	char *filename;
	FILE *file;
	int header_isvalid;
	mp3header header;
	int id3_isvalid;
	id3tag id3;
	int vbr;
	float vbr_average;
	int seconds;
	int frames;
	int badframes;
} mp3info;

static int frequencies[3][4] = {
   {22050,24000,16000,50000},  /* MPEG 2.0 */
   {44100,48000,32000,50000},  /* MPEG 1.0 */
   {11025,12000,8000,50000}    /* MPEG 2.5 */
};

static int mp3info_bitrate[2][3][14] = { 
  { /* MPEG 2.0 */
    {32,48,56,64,80,96,112,128,144,160,176,192,224,256},  /* layer 1 */
    {8,16,24,32,40,48,56,64,80,96,112,128,144,160},       /* layer 2 */
    {8,16,24,32,40,48,56,64,80,96,112,128,144,160}        /* layer 3 */
  },

  { /* MPEG 1.0 */
    {32,64,96,128,160,192,224,256,288,320,352,384,416,448}, /* layer 1 */
    {32,48,56,64,80,96,112,128,160,192,224,256,320,384},    /* layer 2 */
    {32,40,48,56,64,80,96,112,128,160,192,224,256,320}      /* layer 3 */
  }
};

static int frame_size_index[] = {24000, 72000, 72000};
static char *mode_text[] = {"stereo", "joint stereo", "dual channel", "mono"};

static mp3info mp3;
static char mp3buf[8192];
static int chunk_size;

static int header_frequency(mp3header *h)
{
	return frequencies[h->version][h->freq];
}

static char *header_mode(mp3header *h)
{
	return mode_text[h->mode];
}
static int header_bitrate(mp3header *h)
{
	return mp3info_bitrate[h->version & 1][3 - h->layer][h->bitrate - 1];
}


static int frame_length(mp3header *header)
{
	if (header->sync != 0xFFE)
		return 1;
	return (
		frame_size_index[3 - header->layer] 
		* ((header->version & 1) + 1)
		* header_bitrate(header) 
		/ header_frequency(header)
	) + header->padding;
}
static void write_info_str(char *info_str)
{
	int v = mp3.id3_isvalid;

	snprintf(info_str, MMD_INFO_SIZE,
		"audio_file_info1:%i kbit/s (%cbr) %i KHz %s, %d frames\n"
		"audio_file_info2:%s, by %s\n"
		"audio_file_info3:A: %s, Y: %s, C: %s\n",
		header_bitrate(&mp3.header), mp3.vbr? 'v' : 'c',
		header_frequency(&mp3.header) / 1000,
		header_mode(&mp3.header),
		mp3.frames,
		v && *mp3.id3.title? mp3.id3.title : "(title tag not set)",
		v && *mp3.id3.artist? mp3.id3.artist : "(artist tag not set)",
		v && *mp3.id3.album? mp3.id3.album : "(album tag not set)",
		v && *mp3.id3.year? mp3.id3.year : "????",
		v && *mp3.id3.comment? mp3.id3.comment : "(comment tag not set)"
	);
}

static void mp3_close_audio_file(FILE *file)
{
	if (file) {
		clog(DEBUG, "%s: closing audio file\n", __func__);
		fclose(file);
	}
}

static int mp3_write_chunk(FILE* file, char *info_str, long unsigned *chunk_num)
{
	int ret, size = chunk_size, fd = fileno(file);

repeat:
	//clog(DEBUG, "%s: sending %d bytes to fd %d\n", __func__, size, fd);
	ret = write(fd, mp3buf, size);
	if (ret <= 0) {
		if (ret < 0)
			clog(WARNING, "%s: write error (%s), size=%i\n",
				__func__, strerror(errno), size);
		//mp3_stop_stream(file);
		return ret;
	}
	if (ret < size) {
		clog(WARNING, "%s: short write (%i/%i), rewriting\n", __func__, ret, 
			size);
		memmove(mp3buf, mp3buf + ret, size - ret);
		size -= ret;
		goto repeat;
	}
	if (info_str)
		write_info_str(info_str);
	*chunk_num += 1;
	return 1;
}
/* 
 * Remove trailing whitespace from the end of a string 
 */
static char *unpad(char *string)
{
	char *pos = string + strlen(string) - 1;
	while (isspace(pos[0]))
		(pos--)[0] = 0;
	return string;
}

static int compare_headers(mp3header *h1, mp3header *h2)
{
	if ((*(uint*)h1) == (*(uint*)h2))
		return 1;
	if ((h1->version == h2->version) &&
			(h1->layer == h2->layer) &&
			(h1->crc == h2->crc) &&
			(h1->freq == h2->freq) &&
			(h1->mode == h2->mode) &&
			(h1->copyright == h2->copyright) &&
			(h1->original == h2->original) &&
			(h1->emphasis == h2->emphasis)) 
		return 1;
	else
		return 0;
}


static int para_fread(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	size_t res = fread(ptr, size, nmemb, stream);
	if (res == nmemb)
		return size * nmemb;
	if (feof(stream))
		return 0;
	return -1;
}
/* 
 * Get next MP3 frame header.
 * Return codes:
 * positive value: Frame Length of this header
 * 0: No, we did not retrieve a valid frame header
 * < 0: error
 */
static int get_header(FILE *file, mp3header *header)
{
	unsigned char buffer[FRAME_HEADER_SIZE];
	int fl, ret;

	if (!file || !header) {
		clog(ERROR, "file == NULL || header == NULL\n");
		return -1;
	}
	ret = para_fread(buffer, FRAME_HEADER_SIZE, 1, file);
	if (ret < FRAME_HEADER_SIZE) {
		header->sync = 0;
		return ret < 0? -1 : 0;
	}
	header->sync = (((int)buffer[0]<<4) | ((int)(buffer[1]&0xE0)>>4));
	if (buffer[1] & 0x10) 
		header->version = (buffer[1] >> 3) & 1;
	else 
		header->version = 2;
	header->layer = (buffer[1] >> 1) & 3;
	if ((header->sync != 0xFFE) || (header->layer != 1)) {
		header->sync = 0;
		//clog(DEBUG, "%s: header not found\n", __func__);
		return 0;
	}
	//clog(DEBUG, "%s: found header\n", __func__);
	header->crc = buffer[1] & 1;
	header->bitrate = (buffer[2] >> 4) & 0x0F;
	header->freq = (buffer[2] >> 2) & 0x3;
	header->padding = (buffer[2] >>1) & 0x1;
	header->extension = (buffer[2]) & 0x1;
	header->mode = (buffer[3] >> 6) & 0x3;
	header->mode_extension = (buffer[3] >> 4) & 0x3;
	return ((fl = frame_length(header)) >= MIN_FRAME_SIZE ? fl : -1);
}

/*
 * If startpos is >=0, seek to that position and find next header
 */
static int seek_next_header(long startpos) 
{
	int k, l = 0, c, first_len;
	mp3header h, h2;
	long valid_start = 0;
  
	if (startpos >= 0)
		if (fseek(mp3.file, startpos, SEEK_SET) < 0)
			return -1;
	while (1) {
		while ((c = fgetc(mp3.file)) != 255 && (c != EOF))
			; /* nothing */
		if (c != 255) 
			return 0;
		ungetc(c, mp3.file);
		valid_start = ftell(mp3.file);
		first_len = get_header(mp3.file, &h);
		if (first_len <= 0)
			continue;
		if (fseek(mp3.file, first_len - FRAME_HEADER_SIZE, 
				SEEK_CUR) < 0)
			return -1;
		for (k = 1; k < MIN_CONSEC_GOOD_FRAMES; k++) {
			if ((l = get_header(mp3.file, &h2)) <= 0)
				break;
			if (!compare_headers(&h, &h2))
				break;
			fseek(mp3.file,l - FRAME_HEADER_SIZE, SEEK_CUR);
		}
		if (k == MIN_CONSEC_GOOD_FRAMES) {
			fseek(mp3.file,valid_start, SEEK_SET);
			memcpy(&(mp3.header), &h2, sizeof(mp3header));
			mp3.header_isvalid = 1;
			return first_len;
		}
	}
	return 0;
}

static int get_id3(void)
{
	char fbuf[4];

	mp3.id3.title[0] = '\0';
	mp3.id3.artist[0] = '\0';
	mp3.id3.album[0] = '\0';
	mp3.id3.comment[0] = '\0';
	mp3.id3.year[0] = '\0';
	if (fseek(mp3.file, -128, SEEK_END)) {
		clog(ERROR, "%s: Could not read last 128 bytes\n",
			__func__);
		return -1;
	}
	fread(fbuf, 1, 3, mp3.file);
	fbuf[3] = '\0';
	mp3.id3.genre[0] = 255;
	if (!strcmp((const char *)"TAG",(const char *)fbuf)) {
		mp3.id3_isvalid = 1;
		fseek(mp3.file, -125, SEEK_END);
		fread(mp3.id3.title, 1, 30, mp3.file); 
		mp3.id3.title[30] = '\0';
		fread(mp3.id3.artist, 1, 30, mp3.file); 
		mp3.id3.artist[30] = '\0';
		fread(mp3.id3.album, 1, 30, mp3.file);
		mp3.id3.album[30] = '\0';
		fread(mp3.id3.year, 1, 4, mp3.file);	
		mp3.id3.year[4] = '\0';
		fread(mp3.id3.comment, 1, 30, mp3.file);
		mp3.id3.comment[30] = '\0';
		if (mp3.id3.comment[28] == '\0')
			mp3.id3.track[0] = mp3.id3.comment[29];
		fread(mp3.id3.genre, 1, 1, mp3.file);
		unpad(mp3.id3.title);
		unpad(mp3.id3.artist);
		unpad(mp3.id3.album);
		unpad(mp3.id3.year);
		unpad(mp3.id3.comment);
	}
	return 1;
}

static int find_valid_start(FILE *file)
{
	int frame_len;

	if (!file) {
		clog(ERROR, "%s: mp3 == NULL!\n", __func__);
		return -1;
	}
	frame_len = get_header(file, &mp3.header);
	if (frame_len < 0)
		return -1;
	if (frame_len == 0) {
		frame_len = seek_next_header(-1);
		if (frame_len <= 0)
			return frame_len;
	} else if (fseek(file, -FRAME_HEADER_SIZE, SEEK_CUR) < 0)
			return -1;
	if (frame_len <= 1)
		return -1;
	return frame_len;
}

static int read_info(FILE *audio_file)
{
	int count = 0, len;
	int avg = 0, hb, old_hb = -1;
	
	if (get_id3() <= 0)
		clog(INFO, "%s: no id3 tag\n", __func__);
	rewind(mp3.file);
	mp3.vbr = 0;
	while ((len = find_valid_start(audio_file)) > 0) {
		count++;
		hb = header_bitrate(&mp3.header);
		avg = ((count - 1) * avg + hb) / count;
		if (old_hb > 0 && (old_hb != hb))
			mp3.vbr = 1;
		old_hb = hb;
		if (fseek(mp3.file, len, SEEK_CUR) < 0)
			return -1;
	}
	if (!count)
		return -1;
	mp3.frames = count;
	mp3.seconds = count / 38;  /* It's magic, baby! */
	mp3.vbr_average = avg;
	rewind(mp3.file);
	return 1;
}

/*
 * Read mp3 information from audio file
 */
static int mp3_get_file_info(FILE *audio_file, char *info_str, 
	long unsigned *frames, int *seconds)
{
	if (!audio_file) {
		clog(CRIT, "%s: called without audio file", __func__);
		return -1;
	}
	mp3.file = audio_file;
	if (read_info(audio_file) < 0)
		return -1;
	write_info_str(info_str);
	*frames = mp3.frames;
	*seconds = mp3.seconds;
	return 1;
}

static int mp3_reposition_stream(FILE* audio_file, long unsigned frame_no,
		pid_t *pid)
{
	int count = 0, len;
	
	clog(DEBUG, "%s: jmp to frame %i/%i\n", __func__,
		frame_no, mp3.frames);
	//stop_stream_writer(stream_writer);
	rewind(audio_file);
	while ((len = find_valid_start(audio_file)) > 0 && count < frame_no) {
		count++;
		if (fseek(audio_file, len, SEEK_CUR) < 0)
			return -1;
	}
	if (*pid) {
		kill(*pid, SIGTERM);
		*pid = 0;
	}
	if (count == frame_no)
		return 1;
	return -1;
}

static int mp3_read_chunk(FILE *file)
{
	int len = find_valid_start(file);
	
	if (len < 0) {
		clog(ERROR, "%s: invalid frame len (%d)\n", __func__, len);
		return -1;
	}
	chunk_size = para_fread(mp3buf, len, 1, file);
	return chunk_size;
}

void mp3_end_of_input(FILE *in_file, FILE **sw, pid_t* pid)
{
	mp3_close_audio_file(in_file);
}

int mp3_init(void *p)
{
		struct audio_format *af = p;	
		af->get_file_info = mp3_get_file_info,
		af->reposition_stream = mp3_reposition_stream,
		af->read_chunk = mp3_read_chunk;
		af->write_chunk = mp3_write_chunk;
		af->end_of_input = mp3_end_of_input;
		af->close_audio_file = mp3_close_audio_file;
		return 1;
}
